;;; -*- Mode: Lisp; Syntax: ANSI-Common-Lisp; Base: 10 -*-
;;;
;;; collections.lisp -- view Berkeley DBs as Lisp collections
;;; 
;;; Initial version 8/26/2004 by Ben Lee
;;; <blee@common-lisp.net>
;;; 
;;; part of
;;;
;;; Elephant: an object-oriented database for Common Lisp
;;;
;;; Copyright (c) 2004 by Andrew Blumberg and Ben Lee
;;; <ablumberg@common-lisp.net> <blee@common-lisp.net>
;;;
;;; This program is released under the following license
;;; ("GPL").  For differenct licensing terms, contact the
;;; copyright holders.
;;;
;;; This program is free software; you can redistribute it
;;; and/or modify it under the terms of the GNU General
;;; Public License as published by the Free Software
;;; Foundation; either version 2 of the License, or (at
;;; your option) any later version.
;;;
;;; This program is distributed in the hope that it will be
;;; useful, but WITHOUT ANY WARRANTY; without even the
;;; implied warranty of MERCHANTABILITY or FITNESS FOR A
;;; PARTICULAR PURPOSE. See the GNU General Public License
;;; for more details.
;;;
;;; The GNU General Public License can be found in the file
;;; LICENSE which should have been distributed with this
;;; code.  It can also be found at
;;;
;;; http://www.opensource.org/licenses/gpl-license.php
;;;
;;; You should have received a copy of the GNU General
;;; Public License along with this program; if not, write
;;; to the Free Software Foundation, Inc., 59 Temple Place,
;;; Suite 330, Boston, MA 02111-1307 USA
;;;

(in-package "ELEPHANT")

;;; collection types
;;; we're slot-less
(defclass persistent-collection (persistent) 
  ())

;;; btree access
(defclass btree (persistent-collection) ())

(defgeneric get-value (key ht))
(defgeneric (setf get-value) (value key ht))
(defgeneric remove-kv (key ht &key transaction auto-commit))

(defmethod get-value (key (ht btree))
  (declare (optimize (speed 3) (safety 0) (space 3)))
  (buffer-write-int (oid ht) *key-buf*)
  (let* ((key-length (serialize key *key-buf*))
	 (buf (db-get-key-buffered 
	       (controller-db *store-controller*) 
	       (buffer-stream-buffer *key-buf*)
	       key-length)))
    (declare (type fixnum key-length))
    (if buf (values (deserialize buf) T)
	(values nil nil))))

(defmethod (setf get-value) (value key (ht btree))
  (declare (optimize (speed 3) (safety 0)))
  (buffer-write-int (oid ht) *key-buf*)
  (let ((key-length (serialize key *key-buf*))
	(val-length (serialize value *out-buf*)))
    (db-put-buffered (controller-db *store-controller*) 
		     (buffer-stream-buffer *key-buf*) key-length
		     (buffer-stream-buffer *out-buf*) val-length
		     :transaction *current-transaction*
		     :auto-commit *auto-commit*)
    value))

(defmethod remove-kv (key (ht btree)
		      &key (transaction *current-transaction*) 
		      (auto-commit *auto-commit*))
  (declare (optimize (speed 3) (safety 0)))
  (buffer-write-int (oid ht) *key-buf*)
  (let ((key-length (serialize key *key-buf*)))
    (db-delete-buffered (controller-db *store-controller*) 
			(buffer-stream-buffer *key-buf*) key-length
			:transaction transaction
			:auto-commit auto-commit)))
