;;; -*- Mode: Lisp; Syntax: ANSI-Common-Lisp; Base: 10 -*-
;;;
;;; sleepycat.lisp -- FFI interface to Berkeley DB
;;; 
;;; Initial version 8/26/2004 by Ben Lee
;;; <blee@common-lisp.net>
;;; 
;;; part of
;;;
;;; Elephant: an object-oriented database for Common Lisp
;;;
;;; Copyright (c) 2004 by Andrew Blumberg and Ben Lee
;;; <ablumberg@common-lisp.net> <blee@common-lisp.net>
;;;
;;; This program is released under the following license
;;; ("GPL").  For differenct licensing terms, contact the
;;; copyright holders.
;;;
;;; This program is free software; you can redistribute it
;;; and/or modify it under the terms of the GNU General
;;; Public License as published by the Free Software
;;; Foundation; either version 2 of the License, or (at
;;; your option) any later version.
;;;
;;; This program is distributed in the hope that it will be
;;; useful, but WITHOUT ANY WARRANTY; without even the
;;; implied warranty of MERCHANTABILITY or FITNESS FOR A
;;; PARTICULAR PURPOSE. See the GNU General Public License
;;; for more details.
;;;
;;; The GNU General Public License can be found in the file
;;; LICENSE which should have been distributed with this
;;; code.  It can also be found at
;;;
;;; http://www.opensource.org/licenses/gpl-license.php
;;;
;;; You should have received a copy of the GNU General
;;; Public License along with this program; if not, write
;;; to the Free Software Foundation, Inc., 59 Temple Place,
;;; Suite 330, Boston, MA 02111-1307 USA
;;;


(defpackage sleepycat
  (:use common-lisp uffi)
  (:export #:*current-transaction* 
	   #:read-int #:read-uint #:read-float #:read-double 
	   #:write-int #:write-uint #:write-float #:write-double
	   #:offset-char-pointer #:copy-str-to-buf #:copy-bufs #:byte-length
	   #:pointer-int #:pointer-void #:array-or-pointer-char
	   #:db-env-create #:db-env-close #:db-env-open #:db-env-dbremove
	   #:db-env-dbrename #:db-env-remove 
	   #:db-env-set-flags #:db-env-get-flags
	   #:db-create #:db-close #:db-open 
	   #:db-remove #:db-rename #:db-sync #:db-truncate
	   #:db-get-key-buffered #:db-get-buffered #:db-get 
	   #:db-put-buffered #:db-put 
	   #:db-delete-buffered #:db-delete 
	   #:db-transaction-begin #:db-transaction-abort 
	   #:db-transaction-commit #:with-transaction 
	   #:db-transaction-id #:db-env-lock-id #:db-env-lock-id-free
	   #:db-env-lock-get #:db-env-lock-put #:with-lock
	   #:db-env-set-timeout #:db-env-get-timeout
	   #:db-env-set-lock-detect #:db-env-get-lock-detect
	   #:db-error #:db-error-errno
	   #:+NULL-VOID+ #:+NULL-CHAR+
	   #:DB-BTREE #:DB-HASH #:DB-QUEUE #:DB-RECNO #:DB-UNKNOWN 
	   #:DB_KEYEMPTY #:DB_LOCK_DEADLOCK #:DB_LOCK_NOTGRANTED 
	   #:DB_NOTFOUND
	   #:DB-LOCKMODE#NG #:DB-LOCKMODE#READ #:DB-LOCKMODE#WRITE 
	   #:DB-LOCKMODE#wAIT #:DB-LOCKMODE#IWRITE #:DB-LOCKMODE#IREAD
	   #:DB-LOCKMODE#IWR #:DB-LOCKMODE#DIRTY #:DB-LOCKMODE#WWRITE
	   ))

(in-package "SLEEPYCAT")

(eval-when (:compile-toplevel :load-toplevel)
  ;; UFFI
  ;;(asdf:operate 'asdf:load-op :uffi)

  ;; DSO loading - Edit these for your system!

  ;; Under linux you may need to load some kind of pthread
  ;; library.  I can't figure out which is the right one.
  ;; This one worked for me.  There are known issues with
  ;; Red Hat and Berkeley DB, search google.
  #+linux
  (unless 
      (uffi:load-foreign-library "/lib/tls/libpthread.so.0" :module "pthread")
    (error "Couldn't load libpthread!"))

  (unless
      (uffi:load-foreign-library 
       ;; Sleepycat: this works on linux
       #+linux
       "/usr/local/BerkeleyDB.4.2/lib/libdb.so" 
       ;; this works on FreeBSD
       #+(or bsd freebsd)
       "/usr/local/lib/db42/libdb.so" 
       #+darwin
       "/usr/local/BerkeleyDB.4.2/lib/libdb.dylib"
       :module "sleepycat")
    (error "Couldn't load libdb (Sleepycat)!"))

  ;; Libsleepycat.so: edit this
  (unless
      (uffi:load-foreign-library 
       "/usr/local/share/common-lisp/elephant-0.1/libsleepycat.so" 
       :module "libsleepycat")
    (error "Couldn't load libsleepycat!"))

  ;; fini on user editable part

  (def-type pointer-int (* :int))
  (def-type pointer-void :pointer-void)
  (def-foreign-type array-or-pointer-char
      #+allegro (:array :char)
      #+(or cmu sbcl scl openmcl) (* :char))
  (def-type array-or-pointer-char array-or-pointer-char)
  (def-enum DBTYPE ((:BTREE 1) :HASH :QUEUE :RECNO :UNKNOWN))
)

(declaim (inline read-int read-uint read-float read-double 
		 write-int write-uint write-float write-double
		 offset-char-pointer copy-str-to-buf copy-bufs
		 %db-get-key-buffered db-get-key-buffered 
		 %db-get-buffered db-get-buffered db-get 
		 %db-put-buffered db-put-buffered 
		 %db-put db-put 
		 %db-delete db-delete-buffered db-delete 
		 %db-txn-begin db-transaction-begin
		 %db-txn-abort db-transaction-abort
		 %db-txn-commit db-transaction-commit
		 %db-transaction-id
		 flags))

;; Constants and Flags
;; eventually write a macro which generates a custom flag function.

;I don't like the UFFI syntax for enumerations
(defconstant DB-BTREE                 1)
(defconstant DB-HASH                  2)
(defconstant DB-QUEUE                 3)
(defconstant DB-RECNO                 4)
(defconstant DB-UNKNOWN               5)

(defconstant DB_AUTO_COMMIT   #x1000000)
(defconstant DB_JOINENV	      #x0040000)
(defconstant DB_INIT_CDB      #x0001000)
(defconstant DB_INIT_LOCK     #x0002000)
(defconstant DB_INIT_LOG      #x0004000)
(defconstant DB_INIT_MPOOL    #x0008000)
(defconstant DB_INIT_REP      #x0010000)
(defconstant DB_INIT_TXN      #x0020000)
(defconstant DB_RECOVER	      #x0000020)
(defconstant DB_RECOVER_FATAL #x0200000)
(defconstant DB_LOCKDOWN      #x0080000)
(defconstant DB_PRIVATE	      #x0100000)
(defconstant DB_SYSTEM_MEM    #x0400000)
(defconstant DB_THREAD	      #x0000040)
(defconstant DB_FORCE	      #x0000004)
(defconstant DB_DIRTY_READ    #x2000000)
(defconstant DB_CREATE	      #x0000001)
(defconstant DB_EXCL          #x0001000)
(defconstant DB_NOMMAP	      #x0000008)
(defconstant DB_RDONLY	      #x0000010)
(defconstant DB_TRUNCATE      #x0000080)
(defconstant DB_TXN_NOSYNC    #x0000100)
(defconstant DB_TXN_NOWAIT    #x0001000)
(defconstant DB_TXN_SYNC      #x0002000)
(defconstant DB_LOCK_NOWAIT   #x001)

(defconstant DB_GET_BOTH         10)
(defconstant DB_SET_LOCK_TIMEOUT 29)
(defconstant DB_SET_TXN_TIMEOUT  33)

(defconstant DB_KEYEMPTY        -30997)
(defconstant DB_LOCK_DEADLOCK   -30995)
(defconstant DB_LOCK_NOTGRANTED -30994)
(defconstant DB_NOTFOUND        -30990)

(defvar +NULL-VOID+ (make-null-pointer :void))
(defvar +NULL-CHAR+ (make-null-pointer :char))


;; Buffer management / pointer arithmetic

;; Notes: on CMUCL and Allegro: with-cast-pointer +
;; deref-array is faster than FFI + C pointer arithmetic.
;; however pointer arithmetic is usually consing.  OpenMCL
;; supports non-consing pointer arithmentic though.


;; TODO: #+openmcl versions which do macptr arith.

(def-function ("read_int" read-int)
    ((buf array-or-pointer-char)
     (offset :int))
  :returning :int)

(def-function ("read_uint" read-uint)
    ((buf array-or-pointer-char)
     (offset :int))
  :returning :unsigned-int)

(def-function ("read_float" read-float)
    ((buf array-or-pointer-char)
     (offset :int))
  :returning :float)

(def-function ("read_double" read-double)
    ((buf array-or-pointer-char)
     (offset :int))
  :returning :double)

(def-function ("write_int" write-int)
    ((buf array-or-pointer-char)
     (num :int)
     (offset :int))
  :returning :void)

(def-function ("write_uint" write-uint)
    ((buf array-or-pointer-char)
     (num :unsigned-int)
     (offset :int))
  :returning :void)

(def-function ("write_float" write-float)
    ((buf array-or-pointer-char)
     (num :float)
     (offset :int))
  :returning :void)

(def-function ("write_double" write-double)
    ((buf array-or-pointer-char)
     (num :double)
     (offset :int))
  :returning :void)

(def-function ("offset_charp" offset-char-pointer)
    ((p array-or-pointer-char)
     (offset :int))
  :returning array-or-pointer-char)

;; Allegro and Lispworks use 16-bit unicode characters
(defmacro byte-length (s)
  #+(or lispworks (and allegro ics))
  `(let ((l (length ,s))) (+ l l))
  #-(or lispworks (and allegro ics))
  `(length ,s))

;; for copying the bytes of a string to a foreign buffer
;; memcpy is faster than looping!  For Lispworks this causes
;; a string to array conversion, but I don't know how to do
;; any better (fli:replace-foreign-array is promising?)
#-(or cmu sbcl scl openmcl)
(def-function ("copy_buf" copy-str-to-buf)
    ((dest array-or-pointer-char)
     (dest-offset :int)
     (src array-or-pointer-char)
     (src-offset :int)
     (length :int))
  :returning :void)

#+(or cmu sbcl scl)
(def-function ("copy_buf" copy-str-to-buf)
    ((dest array-or-pointer-char)
     (dest-offset :int)
     (src :cstring)
     (src-offset :int)
     (length :int))
  :returning :void)

;; but OpenMCL can't directly pass string bytes.
#+openmcl
(defun copy-str-to-buf (dest dest-offset src src-offset length)
  (declare (optimize (speed 3) (safety 0))
	   (type string src)
	   (type array-or-pointer-char dest)
	   (type fixnum length src-offset dest-offset)
	   (dynamic-extent src dest length))
  (multiple-value-bind (ivector disp)
      (ccl::array-data-and-offset src)
    (ccl::%copy-ivector-to-ptr ivector (+ disp src-offset)
			       dest dest-offset length)))

;; Lisp version, for kicks.  this assumes 8-bit chars!
#+(not (or cmu sbcl scl allegro openmcl lispworks))
(defun copy-str-to-buf (dest dest-offset src src-offset length)
  (declare (optimize (speed 3) (safety 0))
	   (type string src)
	   (type array-or-pointer-char dest)
	   (type fixnum length src-offset dest-offset)
	   (dynamic-extent src dest length))
  (typecase src
    (simple-string
     (loop for i fixnum from 0 below length
	   do
	   (setf (deref-array dest 'array-or-pointer-char (+ i dest-offset)) 
		 (char-code (schar src (+ i src-offset))))))
    (string
     (loop for i fixnum from 0 below length
	   do
	   (setf (deref-array dest 'array-or-pointer-char (+ i dest-offset)) 
		 (char-code (char src (+ i src-offset))))))))

;; For copying two foreign buffers
(def-function ("copy_buf" copy-bufs)
    ((dest array-or-pointer-char)
     (dest-offset :int)
     (src array-or-pointer-char)
     (src-offset :int)
     (length :int))
  :returning :void)    

;; Thread local storage (special variables)

(declaim (type array-or-pointer-char *get-buffer*)
	 (type fixnum *get-buffer-length*))

(defvar *current-transaction* +NULL-VOID+)

(defvar *errno-buffer* (allocate-foreign-object :int 1))

(defvar *get-buffer* (allocate-foreign-object :char 1))
(defvar *get-buffer-length* 0)

(defun resize-get-buffer (length)
  (declare (optimize (speed 3) (safety 0) (space 0))
	   (type fixnum length))
  (if (< length *get-buffer-length*)
      (values *get-buffer* *get-buffer-length*)
      (let ((newlen (max length (* *get-buffer-length* 2))))
	(declare (type fixnum newlen))
	(setq *get-buffer-length* newlen)
	(free-foreign-object *get-buffer*)
	(setq *get-buffer* (allocate-foreign-object :char newlen))
	(values *get-buffer* *get-buffer-length*))))

;; Wrapper macro -- handles errno return values
;; makes flags into keywords
;; makes keyword args, cstring wrappers


(eval-when (:compile-toplevel :load-toplevel)
  (defun make-wrapper-args (args flags keys)
    (if (or flags keys)
	(append (remove-keys (remove 'flags args) keys)
		`(&key ,@flags ,@keys))
	(remove 'flags args)))
  
  (defun remove-keys (args keys)
    (if keys
	(loop for key in keys
	      for kw = (if (atom key) key (first key))
	      for wrapper-args = (remove kw args) then (remove kw wrapper-args)
	      finally (return wrapper-args))
	args))
  
  (defun make-fun-args (args flags)
    (if flags
	(substitute (cons 'flags (symbols-to-kw-pairs flags)) 'flags args)
	(substitute 0 'flags args)))
  
  (defun make-out-args (count)
    (loop for i from 1 to count
	  collect (gensym)))
  
  (defun symbols-to-kw-pairs (symbols)
    (loop for symbol in symbols
	  append (list (intern (symbol-name symbol) "KEYWORD")
		       symbol)))
  
  (defun symbols-to-pairs (symbols)
    (loop for symbol in symbols
	  collect (list symbol symbol)))
  )

(defmacro wrap-errno (names args &key (keys nil) (flags nil)
		      (cstrings nil) (outs 1) (declarations nil)
		      (transaction nil))
  (let ((wname (if (listp names) (first names) names))
	(fname (if (listp names) (second names) 
		   (intern (concatenate 'string "%" (symbol-name names)))))
	(wrapper-args (make-wrapper-args args flags keys))
	(fun-args (make-fun-args args flags))
	(errno (gensym)))
    (if (> outs 1)
	(let ((out-args (make-out-args outs)))
	  `(defun ,wname ,wrapper-args
	    ,@(if declarations (list declarations) (values))
	    (with-cstrings ,(symbols-to-pairs cstrings)
	      (multiple-value-bind ,out-args
		  (,fname ,@fun-args)
		(let ((,errno ,(first out-args)))
		  (declare (type fixnum ,errno))
		  (cond
		    ((= ,errno 0) (values ,@(rest out-args)))
		    ,@(if transaction
			  (list `((or (= ,errno DB_LOCK_DEADLOCK)
				   (= ,errno DB_LOCK_NOTGRANTED))
				  (throw ,transaction ,transaction)))
			  (values))
		    (t (error 'db-error :errno ,errno))))))))
	`(defun ,wname ,wrapper-args
	  ,@(if declarations (list declarations) (values))
	  (with-cstrings ,(symbols-to-pairs cstrings)
	    (let ((,errno (,fname ,@fun-args)))
	      (declare (type fixnum ,errno))
	      (cond 
		((= ,errno 0) nil)
		,@(if transaction
		      (list `((or (= ,errno DB_LOCK_DEADLOCK)
			       (= ,errno DB_LOCK_NOTGRANTED))
			      (throw ,transaction ,transaction)))
		      (values))
		(t (error 'db-error :errno ,errno)))))))))


;; Environment

(def-function ("db_env_cr" %db-env-create)
    ((flags :unsigned-int)
     (errno :int :out))
  :returning :pointer-void)

(defun db-env-create ()
  (multiple-value-bind (env errno)
      (%db-env-create 0)
    (declare (type fixnum errno))
    (if (= errno 0)
	env
	(error 'db-error :errno errno))))
	     
(def-function ("db_env_close" %db-env-close)
    ((dbenvp :pointer-void)
     (flags :unsigned-int))
  :returning :int)

(wrap-errno db-env-close (dbenvp flags))

(def-function ("db_env_open" %db-env-open)
    ((dbenvp :pointer-void)
     (home :cstring)
     (flags :unsigned-int)
     (mode :int))
  :returning :int)

(wrap-errno db-env-open (dbenvp home flags mode)
	    :flags (joinenv init-cdb init-lock init-log 
			    init-mpool init-rep init-txn
			    recover recover-fatal create
			    lockdown private system-mem thread)
	    :keys ((mode #o640))
	    :cstrings (home))

(def-function ("db_env_dbremove" %db-env-dbremove)
    ((env :pointer-void)
     (txn :pointer-void)
     (file :cstring)
     (database :cstring)
     (flags :unsigned-int))
  :returning :int)

(wrap-errno db-env-dbremove (env transaction file database flags) 
	    :flags (auto-commit)
	    :keys ((transaction *current-transaction*)
		   (database +NULL-CHAR+))
	    :cstrings (file database)
	    :transaction transaction)

(def-function ("db_env_dbrename" %db-env-dbrename)
    ((env :pointer-void)
     (txn :pointer-void)
     (file :cstring)
     (database :cstring)
     (newname :cstring)
     (flags :unsigned-int))
  :returning :int)

(wrap-errno db-env-dbrename (env transaction file database newname flags) 
	    :flags (auto-commit)
	    :keys ((transaction *current-transaction*)
		   (database +NULL-CHAR+))
	    :cstrings (file database newname)
	    :transaction transaction)

(def-function ("db_env_remove" %db-env-remove)
    ((env :pointer-void)
     (home :cstring)
     (flags :unsigned-int))
  :returning :int)

(wrap-errno db-env-remove (env home flags) :flags (force)
	    :cstrings (home))

(def-function ("db_env_set_flags" %db-env-set-flags)
    ((env :pointer-void)
     (flags :unsigned-int)
     (onoff :int))
  :returning :int)

(wrap-errno db-env-set-flags (env flags onoff)
	    :flags (auto-commit nommap txn-nosync))

(def-function ("db_env_get_flags" %db-env-get-flags)
    ((env :pointer-void)
     (flags :unsigned-int :out))
  :returning :int)

(wrap-errno db-env-get-flags (env) :outs 2)
	   

;; Database

(def-function ("db_cr" %db-create)
    ((dbenv :pointer-void)
     (flags :unsigned-int)
     (errno :int :out))
  :returning :pointer-void)
	  
(defun db-create (&optional (dbenv +NULL-VOID+))
  (multiple-value-bind (db errno)
      (%db-create dbenv 0)
    (declare (type fixnum errno))
    (if (= errno 0) 
	db
	(error 'db-error :errno errno))))

(def-function ("db_close" %db-close)
    ((db :pointer-void)
     (flags :unsigned-int))
  :returning :int)

(wrap-errno db-close (db flags))

(def-function ("db_open" %db-open)
    ((db :pointer-void)
     (txn :pointer-void)
     (file :cstring)
     (database :cstring)
     (type DBTYPE)
     (flags :unsigned-int)
     (mode :int))
  :returning :int)

(wrap-errno db-open (db transaction file database type flags mode)
	    :flags (auto-commit create dirty-read excl nommap 
				rdonly thread truncate)
	    :keys ((transaction *current-transaction*)
		   (file +NULL-CHAR+)
		   (database +NULL-CHAR+)
		   (type DB-UNKNOWN)
		   (mode #o640))
	    :cstrings (file database)
	    :transaction transaction)
		
(def-function ("db_remove" %db-remove)
    ((db :pointer-void)
     (file :cstring)
     (database :cstring)
     (flags :unsigned-int))
  :returning :int)

(wrap-errno db-remove (db file database flags)
	    :keys ((database +NULL-CHAR+))
	    :cstrings (file database))

(def-function ("db_rename" %db-rename)
    ((db :pointer-void)
     (file :cstring)
     (database :cstring)
     (newname :cstring)
     (flags :unsigned-int))
  :returning :int)

(wrap-errno db-rename (db file database newname flags)
	    :keys ((database +NULL-CHAR+))
	    :cstrings (file database newname))

(def-function ("db_sync" %db-sync)
    ((db :pointer-void)
     (flags :unsigned-int))
  :returning :int)

(wrap-errno db-sync (db flags))

(def-function ("db_truncate" %db-truncate)
    ((db :pointer-void)
     (txn :pointer-void)
     (count :unsigned-int :out)
     (flags :unsigned-int))
  :returning :int)

(wrap-errno db-truncate (db transaction flags) :flags (auto-commit) 
	    :keys ((transaction *current-transaction*)) :outs 2
	    :transaction transaction)

;; Accessors

(def-function ("db_get_raw" %db-get-key-buffered)
    ((db :pointer-void)
     (txn :pointer-void)
     (key array-or-pointer-char)
     (key-length :unsigned-int)
     (buffer array-or-pointer-char)
     (buffer-length :unsigned-int)
     (flags :unsigned-int)
     (result-length :unsigned-int :out))
  :returning :int)

(defun db-get-key-buffered (db key-buffer key-length &key
			    (transaction *current-transaction*)
			    auto-commit get-both dirty-read)
  (declare (optimize (speed 3) (safety 0) (space 0))
	   (type pointer-void db transaction)
	   (type array-or-pointer-char key-buffer)
	   (type fixnum key-length)
	   (type boolean auto-commit get-both dirty-read))
  (loop 
   do
   (multiple-value-bind (errno result-length)
       (%db-get-key-buffered db transaction key-buffer key-length 
			     *get-buffer* *get-buffer-length*
			     (flags :auto-commit auto-commit
				    :get-both get-both
				    :dirty-read dirty-read))
     (declare (type fixnum result-length errno))
     (if (<= result-length *get-buffer-length*)
	 (cond 
	   ((= errno 0)
	    (return-from db-get-key-buffered 
	      (the (values array-or-pointer-char fixnum)
		(values *get-buffer* result-length))))
	   ((or (= errno DB_NOTFOUND) (= errno DB_KEYEMPTY))
		(return-from db-get-key-buffered 
		  (the (values null fixnum) (values nil 0))))
	   ((or (= errno DB_LOCK_DEADLOCK) (= errno DB_LOCK_NOTGRANTED))
	    (throw transaction transaction))
	   (t (error 'db-error :errno errno)))
	 (resize-get-buffer result-length)))))

(def-function ("db_get_raw" %db-get-buffered)
    ((db :pointer-void)
     (txn :pointer-void)
     (key :cstring)
     (key-length :unsigned-int)
     (buffer array-or-pointer-char)
     (buffer-length :unsigned-int)
     (flags :unsigned-int)
     (result-length :unsigned-int :out))
  :returning :int)

(defun db-get-buffered (db key &key
			(key-length (length key))
			(transaction *current-transaction*)
			auto-commit get-both dirty-read)
  (declare (optimize (speed 3) (safety 0) (space 0))
	   (type pointer-void db transaction)
	   (type string key)
	   (type fixnum key-length)
	   (type boolean auto-commit get-both dirty-read))
  (with-cstring (k key)
    (loop 
     do
     (multiple-value-bind (errno result-length)
	 (%db-get-buffered db transaction k key-length 
			   *get-buffer* *get-buffer-length*
			   (flags :auto-commit auto-commit
				  :get-both get-both
				  :dirty-read dirty-read))
       (declare (type fixnum result-length errno))
       (if (<= result-length *get-buffer-length*)
	   (cond 
	     ((= errno 0)
	      (return-from db-get-buffered 
		(the (values array-or-pointer-char fixnum)
		  (values *get-buffer* result-length))))
	     ((or (= errno DB_NOTFOUND) (= errno DB_KEYEMPTY))
		  (return-from db-get-buffered 
		    (the (values null fixnum) (values nil 0))))
	     ((or (= errno DB_LOCK_DEADLOCK) (= errno DB_LOCK_NOTGRANTED))
	      (throw transaction transaction))
	     (t (error 'db-error :errno errno)))
	   (resize-get-buffer result-length))))))

(defun db-get (db key &key (key-length (length key))
	       (transaction *current-transaction*)
	       auto-commit get-both dirty-read)
  (declare (optimize (speed 3) (safety 0) (space 0))
	   (type pointer-void db transaction)
	   (type string key)
	   (type fixnum key-length)
	   (type boolean auto-commit get-both dirty-read))
  (with-cstring (k key)
    (loop 
     do
     (multiple-value-bind (errno result-length)
	 (%db-get-buffered db transaction k key-length 
			   *get-buffer* *get-buffer-length*
			   (flags :auto-commit auto-commit
				  :get-both get-both
				  :dirty-read dirty-read))
       (declare (type fixnum result-length errno))
       (if (<= result-length *get-buffer-length*)
	   (cond
	     ((= errno 0)
	      (return-from db-get
		(convert-from-foreign-string *get-buffer* 
					     :length result-length
					     :null-terminated-p nil)))
	     ((or (= errno DB_NOTFOUND) (= errno DB_KEYEMPTY))
	      (return-from db-get nil))
	     ((or (= errno DB_LOCK_DEADLOCK) (= errno DB_LOCK_NOTGRANTED))
	      (throw transaction transaction))
	     (t (error 'db-error :errno errno)))
	   (resize-get-buffer result-length))))))

(def-function ("db_put_raw" %db-put-buffered)
    ((db :pointer-void)
     (txn :pointer-void)
     (key array-or-pointer-char)
     (key-length :unsigned-int)
     (datum array-or-pointer-char)
     (datum-length :unsigned-int)
     (flags :unsigned-int))
  :returning :int)

(wrap-errno db-put-buffered (db transaction key key-length 
				datum datum-length flags)
	    :flags (auto-commit)
	    :keys ((transaction *current-transaction*))
	    :declarations (declare (optimize (speed 3) (safety 0) (space 0))
				   (type pointer-void db transaction)
				   (type array-or-pointer-char key datum)
				   (type fixnum key-length datum-length)
				   (type boolean auto-commit))
	    :transaction transaction)

(def-function ("db_put_raw" %db-put)
    ((db :pointer-void)
     (txn :pointer-void)
     (key :cstring)
     (key-length :unsigned-int)
     (datum :cstring)
     (datum-length :unsigned-int)
     (flags :unsigned-int))
  :returning :int)

(wrap-errno db-put (db transaction key key-length datum datum-length flags)
	    :flags (auto-commit)
	    :keys ((key-length (length key))
		   (datum-length (length datum))
		   (transaction *current-transaction*))
	    :cstrings (key datum)
	    :declarations (declare (optimize (speed 3) (safety 0) (space 0))
				   (type pointer-void db transaction)
				   (type string key datum)
				   (type fixnum key-length datum-length)
				   (type boolean auto-commit))
	    :transaction transaction)

(def-function ("db_del" %db-delete-buffered)
    ((db :pointer-void)
     (txn :pointer-void)
     (key array-or-pointer-char)
     (key-length :unsigned-int)
     (flags :unsigned-int))
  :returning :int)

(defun db-delete-buffered  (db key key-length &key auto-commit 
			    (transaction *current-transaction*))
  (declare (optimize (speed 3) (safety 0) (space 0))
	   (type pointer-void db transaction) (type array-or-pointer-char key)
	   (type fixnum key-length) (type boolean auto-commit))
  (let ((errno (%db-delete-buffered db transaction
				    key key-length 
				    (flags :auto-commit auto-commit))))
    (declare (type fixnum errno))
    (cond ((= errno 0) t)
	  ((or (= errno DB_NOTFOUND) 
	       (= errno DB_KEYEMPTY))
	   nil)
	  ((or (= errno DB_LOCK_DEADLOCK)
	       (= errno DB_LOCK_NOTGRANTED))
	   (throw transaction transaction))
	  (t (error 'db-error :errno errno)))))

(def-function ("db_del" %db-delete)
    ((db :pointer-void)
     (txn :pointer-void)
     (key :cstring)
     (key-length :unsigned-int)
     (flags :unsigned-int))
  :returning :int)

(defun db-delete (db key &key auto-commit (key-length (length key))
		  (transaction *current-transaction*))
  (declare (optimize (speed 3) (safety 0) (space 0))
	   (type pointer-void db transaction) (type string key)
	   (type fixnum key-length) (type boolean auto-commit))
  (with-cstrings ((key key))
    (let ((errno
	   (%db-delete db transaction key
		       key-length (flags :auto-commit auto-commit))))
      (declare (type fixnum errno))
      (cond ((= errno 0) nil)
	    ((or (= errno DB_NOTFOUND) 
		 (= errno DB_KEYEMPTY))
	     nil)
	    ((or (= errno DB_LOCK_DEADLOCK)
		 (= errno DB_LOCK_NOTGRANTED))
	     (throw transaction transaction))
	    (t (error 'db-error :errno errno))))))

;; Transactions

(def-function ("db_txn_begin" %db-txn-begin)
    ((env :pointer-void)
     (parent :pointer-void)
     (flags :unsigned-int)
     (errno (* :int)))
  :returning :pointer-void)

(defun db-transaction-begin (env &key (parent *current-transaction*)
			     dirty-read txn-nosync txn-nowait
			     txn-sync)
  (declare (optimize (speed 3) (safety 0) (space 0))
	   (type pointer-void env parent)
	   (type boolean dirty-read txn-nosync txn-nowait
		 txn-sync)
	   (type pointer-int *errno-buffer*))
  (let* ((txn
	  (%db-txn-begin env parent
			 (flags :dirty-read dirty-read
				:txn-nosync txn-nosync
				:txn-nowait txn-nowait
				:txn-sync txn-sync)
			 *errno-buffer*))
	 (errno (deref-array *errno-buffer* '(:array :int) 0)))
    (declare (type pointer-void txn)
	     (type fixnum errno))
    (if (= errno 0) 
	txn
	(error 'db-error :errno errno))))

(def-function ("db_txn_abort" %db-txn-abort)
    ((txn :pointer-void))
  :returning :int)

(wrap-errno (db-transaction-abort %db-txn-abort) (transaction)
	    :keys ((transaction *current-transaction*))
	    :declarations (declare (optimize (speed 3) (safety 0) (space 0))
				   (type pointer-void transaction)))

(def-function ("db_txn_commit" %db-txn-commit)
    ((txn :pointer-void)
     (flags :unsigned-int))
  :returning :int)

(wrap-errno (db-transaction-commit %db-txn-commit) (transaction flags)
	    :keys ((transaction *current-transaction*))
	    :flags (txn-nosync txn-sync)
	    :declarations (declare (optimize (speed 3) (safety 0) (space 0))
				   (type pointer-void transaction)
				   (type boolean txn-nosync txn-sync)))

(defmacro with-transaction ((&key transaction environment
				  (parent '*current-transaction*)
				  (retries 100)
				  dirty-read txn-nosync
				  txn-nowait txn-sync)
			    &body body)
  (let ((txn (if transaction transaction (gensym)))
	(count (gensym))
	(result (gensym))
	(success (gensym)))
    `(loop for ,count fixnum from 1 to ,retries
	   for ,txn of-type pointer-void = 
	   (db-transaction-begin ,environment
				 :parent ,parent
				 :dirty-read ,dirty-read
				 :txn-nosync ,txn-nosync
				 :txn-nowait ,txn-nowait
				 :txn-sync ,txn-sync)
	   for ,success of-type boolean = nil
           for ,result = 	 
           (let ((*current-transaction* ,txn))
	     (catch ,txn
	       (unwind-protect
		    (prog1 (progn ,@body)
		      (setq ,success t)
		      (db-transaction-commit :transaction ,txn 
					     :txn-nosync ,txn-nosync
					     :txn-sync ,txn-sync))
		 (unless ,success 
		   (db-transaction-abort :transaction ,txn)))))
           do
	   (unless (and (eq ,result ,txn) (not ,success))
	     (return ,result))
           finally (error "Too many retries"))))

;; this is code for a non-consing with-transaction. which
;; doesn't work in the (globally t) case (e.g. setting
;; *current-transaction*.)

;    #+cmu
;    `(alien:with-alien ((,txn (* t)
;			 (%db-txn-begin 
;			  ,environment ,parent
;			  (flags :dirty-read ,dirty-read
;				 :txn-nosync ,txn-nosync
;				 :txn-nowait ,txn-nowait
;				 :txn-sync ,txn-sync)
;			  *errno-buffer*)))
;      (let ((,success nil)
;	    ,@(if globally `((*current-transaction* ,txn)) (values)))
;	(declare (type pointer-void *current-transaction*)
;		 (dynamic-extent *current-transaction*))
;	(unwind-protect
;	     (prog1 (progn ,@body)
;	       (setq ,success t)
;	       (%db-txn-commit ,txn
;			       (flags :txn-nosync ,txn-nosync
;				      :txn-sync ,txn-sync)))
;	  (unless ,success (%db-txn-abort ,txn)))))))
	       

;; Locks and timeouts


(def-enum DB-LOCKOP ((:DUMP 0) :GET :GET-TIMEOUT :INHERIT 
		     :PUT :PUT-ALL :PUT-OBJ :PUT-READ
		     :TIMEOUT :TRADE :UPGRADE-WRITE))

(def-enum DB-LOCKMODE ((:NG 0) :READ :WRITE :WAIT 
		       :IWRITE :IREAD :IWR :DIRTY :WWRITE))

(def-struct DB-LOCK
    (off :unsigned-int)
  (ndx :unsigned-int)
  (gen :unsigned-int)
  (mode DB-LOCKMODE))

#+openmcl
(ccl:def-foreign-type DB-LOCK (:struct DB-LOCK))

(def-struct DB-LOCKREQ
    (op DB-LOCKOP)
  (mode DB-LOCKMODE)
  (timeout :unsigned-int)
  (obj (:array :char))
  (lock (* DB-LOCK)))

#+openmcl
(ccl:def-foreign-type DB-LOCKREQ (:struct DB-LOCKREQ))

(def-function ("db_txn_id" %db-transaction-id)
    ((transaction :pointer-void))
  :returning :unsigned-int)

(defun db-transaction-id (&optional (transaction *current-transaction*))
  (%db-transaction-id transaction))

(def-function ("db_env_lock_id" %db-env-lock-id)
    ((env :pointer-void)
     (id :unsigned-int :out))
  :returning :int)

(wrap-errno db-env-lock-id (env) :outs 2)


(def-function ("db_env_lock_id_free" %db-env-lock-id-free)
    ((env :pointer-void)
     (id :unsigned-int))
  :returning :int)

(wrap-errno db-env-lock-id-free (env id))

(def-function ("db_env_lock_get" %db-env-lock-get)
    ((env :pointer-void)
     (locker :unsigned-int)
     (flags :unsigned-int)
     (object array-or-pointer-char)
     (object-length :unsigned-int)
     (lock-mode DB-LOCKMODE)
     (lock (* DB-LOCK)))
  :returning :int)

(wrap-errno db-env-lock-get (env locker flags object object-length 
				 lock-mode lock)
	    :flags (lock-nowait))

(def-function ("db_env_lock_put" %db-env-lock-put)
    ((env :pointer-void)
     (lock (* DB-LOCK)))
  :returning :int)

(wrap-errno db-env-lock-put (env lock))

(defmacro with-lock ((env locker object object-length 
			  &key (lock-mode DB-LOCKMODE#WRITE)
			  lock-nowait)
		     &body body)
  (let ((lock (gensym))
	(locked (gensym)))
  `(with-foreign-object (,lock 'DB-LOCK)
    (let ((,locked nil))
      (unwind-protect
	   (progn 
	     (db-env-lock-get ,env ,locker ,object ,object-length ,lock-mode
			      ,lock :lock-nowait ,lock-nowait)
	     (setq ,locked T)
	     ,@body)
	(when ,locked (db-env-lock-put ,env ,lock)))))))      
     
(def-function ("db_env_lock_vec" %db-env-lock-vec)
    ((env :pointer-void)
     (locker :unsigned-int)
     (flags :unsigned-int)
     (list (:array DB-LOCKREQ))
     (nlist :int)
     (elistp (* (* DB-LOCKREQ))))
  :returning :int)

(def-function ("db_env_set_timeout" %db-env-set-timeout)
    ((env :pointer-void)
     (timeout :unsigned-int)
     (flags :unsigned-int))
  :returning :int)

(wrap-errno db-env-set-timeout (env timeout flags)
	    :flags (set-lock-timeout set-transaction-timeout))

(def-function ("db_env_get_timeout" %db-env-get-timeout)
    ((env :pointer-void)
     (timeout :unsigned-int :out)
     (flags :unsigned-int))
  :returning :int)

(wrap-errno db-env-get-timeout (env flags) :outs 2
	    :flags (set-lock-timeout set-transaction-timeout))

(defconstant DB_LOCK_DEFAULT	     1)
(defconstant DB_LOCK_EXPIRE	     2)
(defconstant DB_LOCK_MAXLOCKS        3)
(defconstant DB_LOCK_MINLOCKS        4)
(defconstant DB_LOCK_MINWRITE        5)
(defconstant DB_LOCK_OLDEST	     6)
(defconstant DB_LOCK_RANDOM	     7)
(defconstant DB_LOCK_YOUNGEST        8)

(def-function ("db_env_set_lk_detect" %db-env-set-lock-detect)
    ((env :pointer-void)
     (detect :unsigned-int))
  :returning :int)

(wrap-errno db-env-set-lock-detect (env detect))

(def-function ("db_env_get_lk_detect" %db-env-get-lock-detect)
    ((env :pointer-void)
     (detect :unsigned-int :out))
  :returning :int)

(wrap-errno db-env-get-lock-detect (env) :outs 2)

(def-function ("db_env_lock_detect" %db-env-lock-detect)
    ((env :pointer-void)
     (flags :unsigned-int)
     (atype :unsigned-int)
     (aborted :int :out))
  :returning :int)

(wrap-errno db-env-lock-detect (env flags atype) :outs 2)

;; Poor man's counters

(def-function ("next_counter" %next-counter)
    ((env :pointer-void)
     (db :pointer-void)
     (parent :pointer-void)
     (key array-or-pointer-char)
     (key-length :unsigned-int)
     (lockid array-or-pointer-char)
     (lockid-length :unsigned-int))
  :returning :int)

(defun next-counter (env db parent key key-length lockid lockid-length)
  (let ((ret (%next-counter env db parent key key-length lockid lockid-length)))
    (if (< ret 0)
	(error 'db-error :errno ret)
	ret)))

;; Misc

(defun flags (&key
	      auto-commit
	      joinenv 
	      init-cdb
	      init-lock
	      init-log
	      init-mpool
	      init-rep
	      init-txn
	      recover
	      recover-fatal
	      lockdown
	      private
	      system-mem
	      thread
	      force
	      get-both
	      dirty-read
	      create
	      excl
	      nommap
	      rdonly
	      truncate
	      txn-nosync
	      txn-nowait
	      txn-sync
	      set-lock-timeout
	      set-transaction-timeout
	      lock-nowait)
  (let ((flags 0))
    (declare (optimize (speed 3) (safety 0) (space 0))
	     (type (unsigned-byte 32) flags)
	     (type boolean auto-commit joinenv init-cdb init-lock
		   init-log init-mpool init-rep init-txn
		   recover recover-fatal lockdown private
		   system-mem thread force get-both
		   dirty-read create excl nommap rdonly
		   truncate txn-nosync txn-nowait
		   set-lock-timeout set-transaction-timeout))
    (when auto-commit (setq flags (logior flags DB_AUTO_COMMIT)))
    (when joinenv (setq flags (logior flags DB_JOINENV)))
    (when init-cdb (setq flags (logior flags DB_INIT_CDB)))
    (when init-lock (setq flags (logior flags DB_INIT_LOCK)))
    (when init-log (setq flags (logior flags DB_INIT_LOG)))
    (when init-mpool (setq flags (logior flags DB_INIT_MPOOL)))
    (when init-rep (setq flags (logior flags DB_INIT_REP)))
    (when init-txn (setq flags (logior flags DB_INIT_TXN)))
    (when recover (setq flags (logior flags DB_RECOVER)))
    (when recover-fatal (setq flags (logior flags DB_RECOVER_FATAL)))
    (when lockdown (setq flags (logior flags DB_LOCKDOWN)))
    (when private (setq flags (logior flags DB_PRIVATE)))
    (when system-mem (setq flags (logior flags DB_SYSTEM_MEM)))
    (when thread (setq flags (logior flags DB_THREAD)))
    (when force (setq flags (logior flags DB_FORCE)))
    (when get-both (setq flags (logior flags DB_GET_BOTH)))
    (when dirty-read (setq flags (logior flags DB_DIRTY_READ)))
    (when create (setq flags (logior flags DB_CREATE)))
    (when excl (setq flags (logior flags DB_EXCL)))
    (when nommap (setq flags (logior flags DB_NOMMAP)))
    (when rdonly (setq flags (logior flags DB_RDONLY)))
    (when truncate (setq flags (logior flags DB_TRUNCATE)))
    (when txn-nosync (setq flags (logior flags DB_TXN_NOSYNC)))
    (when txn-nowait (setq flags (logior flags DB_TXN_NOWAIT)))
    (when txn-sync (setq flags (logior flags DB_TXN_SYNC)))
    (when set-lock-timeout (setq flags (logior flags DB_SET_LOCK_TIMEOUT)))
    (when set-transaction-timeout (setq flags (logior flags DB_SET_TXN_TIMEOUT)))
    (when lock-nowait (setq flags (logior flags DB_LOCK_NOWAIT)))
    flags))

;; Errors

(def-function ("db_strerr" %db-strerror)
    ((error :int))
  :returning :cstring)

(defun db-strerror (errno)
  (convert-from-cstring (%db-strerror errno)))

(define-condition db-error (error) 
  ((errno :type fixnum :initarg :errno :reader db-error-errno))
  (:report
   (lambda (condition stream)
     (declare (type db-error condition) (type stream stream))
     (format stream "Berkeley DB error: ~A"
	     (db-strerror (db-error-errno condition))))))
